#include "llvm/Pass.h"
#include "llvm/IR/IRBuilder.h"
#include "llvm/IR/Module.h"
#include "llvm/IR/Function.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/IR/InstrTypes.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"

/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   taken from https://github.com/quarkslab/llvm-dev-meeting-tutorial-2015/" 
   ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

using namespace llvm;

// anonymous namespace -> avoid exporting unneeded symbols
namespace {

// A pass that perform a simple instruction substitution
// see http://llvm.org/docs/WritingAnLLVMPass.html#the-basicblockpass-class
class MBA : public BasicBlockPass {

public:
  // The address of this static is used to uniquely identify this pass in the
  // pass registry. The PassManager relies on this address to find instance of
  // analyses passes and build dependencies on demand.
  // The value does not matter.
  static char ID;

  MBA() : BasicBlockPass(ID)
  {}

  // Called for each basic block of the module
  // Rely on the equality: a + b == (a ^ b) + 2 * (a & b)
  bool runOnBasicBlock(BasicBlock &BB) override {
    bool modified = false;

    // Can't use a for-range loop because we want to delete the instruction from
    // the list we're iterating when replacing it.
    for (auto IIT = BB.begin(), IE = BB.end(); IIT != IE; ++IIT) {
      Instruction &Inst = *IIT;

      // not regular C++ a dynamic_cast!
      // see http://llvm.org/docs/ProgrammersManual.html#the-isa-cast-and-dyn-cast-templates
      auto *BinOp = dyn_cast<BinaryOperator>(&Inst);

      if (!BinOp)
        // The instruction is not a binary operator, we don't handle it.
        continue;

      unsigned Opcode = BinOp->getOpcode();
      if (!(Opcode == Instruction::Add) || !(BinOp->getType()->isIntegerTy()))
	continue;
      // Only handle integer add.
      // Note instead of doing a dyn_cast to a BinaryOperator above, we could
      // have done directly a dyn_cast to AddOperator, but this way seems more
      // effective to later add other operator.
      
      
      
      // The IRBuilder helps you inserting instructions in a clean and
      // fast way
      // see
      // http://llvm.org/docs/ProgrammersManual.html#creating-and-inserting-new-instructions
      IRBuilder<> Builder(BinOp);
      
      Value *NewValue = Builder.CreateAdd(
					  Builder.CreateXor(BinOp->getOperand(0),
							    BinOp->getOperand(1)),
					  Builder.CreateShl(
							    Builder.CreateAnd(
									      BinOp->getOperand(0),
									      BinOp->getOperand(1)),
							    ConstantInt::get(BinOp->getType(), 1))
					  );
      
      
      // ReplaceInstWithValue basically does this (`IIT' is passed by reference):
      // IIT->replaceAllUsesWith(NewValue);
      // IIT = BB.getInstList.erase(IIT);
      //
      // `erase' returns a valid iterator of the instruction before the
      // one that has been erased. This keeps iterators valid.
      //
      // see also
      // http://llvm.org/docs/ProgrammersManual.html#replacing-an-instruction-with-another-value
      ReplaceInstWithValue(BB.getInstList(),
			   IIT, NewValue);
      modified = true;
      
    }
    return modified;
  }
};
}

// pass registration is done through the constructor of static objects...

/* opt pass registration
 */
char MBA::ID = 0;
static RegisterPass<MBA>
    X("mba",  // the option name -> -mba
      "Mixed Boolean Arithmetic Substitution", // option description
      true, // true as we don't modify the CFG
      false // true if we're writing an analysis
      );

/* clang pass registration (optional)
 */
#include "llvm/IR/LegacyPassManager.h"
#include "llvm/Transforms/IPO/PassManagerBuilder.h"

static void registerClangPass(const PassManagerBuilder &,
                              legacy::PassManagerBase &PM)
{ PM.add(new MBA()); }

// Note the registration point, clang offers several insertion points where you
// can insert your pass.
static RegisterStandardPasses RegisterClangPass
    (PassManagerBuilder::EP_EarlyAsPossible, registerClangPass);

